@@ -1,15 +1,13 @@
 """ Callbacks for the manager class """
-import torch
 import subprocess
 from typing import List
-from .base import Callback
-from torchfitter.conventions import ParamsDict
+
+import torch
+from rich.progress import BarColumn, Progress, TimeRemainingColumn
 from torch.optim.swa_utils import AveragedModel
-from rich.progress import (
-    Progress,
-    BarColumn,
-    TimeRemainingColumn,
-)
+
+from torchfitter.conventions import ParamsDict
+from torchfitter.callbacks.base import Callback
 
 
 class EarlyStopping(Callback):
@@ -35,17 +33,20 @@ class EarlyStopping(Callback):
         self.patience = patience
         self.load_best = load_best
 
-        self.log_name = 'EarlyStopping'
+        self.log_name = "EarlyStopping"
 
     def __repr__(self) -> str:
-        return f"EarlyStopping(patience={self.patience}, load_best={self.load_best})"
+        return (
+            f"EarlyStopping(patience={self.patience}, "
+            f"load_best={self.load_best})"
+        )
 
-    def on_fit_start(self, params_dict):
+    def on_fit_start(self, params_dict: dict) -> None:
         self.wait = 0
         self.stopped_epoch = 0
         self.best = float("inf")
 
-    def on_epoch_end(self, params_dict):
+    def on_epoch_end(self, params_dict: dict) -> None:
         current_loss = params_dict[ParamsDict.VAL_LOSS]
         epoch_number = params_dict[ParamsDict.EPOCH_NUMBER]
         model = params_dict[ParamsDict.MODEL]
@@ -69,7 +70,7 @@ class EarlyStopping(Callback):
                     model.load_state_dict(best_params)
                     self.logger.info("Best observed parameters loaded.")
 
-    def on_fit_end(self, params_dict):
+    def on_fit_end(self, params_dict: dict) -> None:
         if self.stopped_epoch > 0:
             self.logger.info(
                 f"Early stopping applied at epoch: {self.stopped_epoch}"
@@ -96,18 +97,24 @@ class LoggerCallback(Callback):
         Number of decimals in the numbers.
     """
 
-    def __init__(self, update_step, precision=2):
+    def __init__(self, update_step: int, precision: int = 2):
         super(LoggerCallback, self).__init__()
         self.update_step = update_step
         self.prec = precision
 
-        self.log_name = 'LoggerCallback'
+        self.log_name = "LoggerCallback"
 
-    def on_fit_start(self, params_dict):
+    def __repr__(self) -> str:
+        return (
+            f"LoggerCallback(update_step={self.update_step}, "
+            f"precision={self.prec})"
+        )
+
+    def on_fit_start(self, params_dict: dict) -> None:
         dev = params_dict[ParamsDict.DEVICE]
         self.logger.info(f"Starting training process on {dev}")
 
-    def on_epoch_end(self, params_dict) -> None:
+    def on_epoch_end(self, params_dict: dict) -> None:
         epoch_number = params_dict[ParamsDict.EPOCH_NUMBER]
         total_epochs = params_dict[ParamsDict.TOTAL_EPOCHS]
         val_loss = params_dict[ParamsDict.VAL_LOSS]
@@ -124,7 +131,7 @@ class LoggerCallback(Callback):
         if epoch_number % self.update_step == 0 or epoch_number == 1:
             self.logger.info(msg)
 
-    def on_fit_end(self, params_dict):
+    def on_fit_end(self, params_dict: dict) -> None:
         total_time = params_dict[ParamsDict.TOTAL_TIME]
         # final message
         self.logger.info(
@@ -164,7 +171,12 @@ class LearningRateScheduler(Callback):
     >>> from torch.optim.lr_scheduler import ReduceLROnPlateau
     >>> from torchfitter.callbacks import LearningRateScheduler
     >>> sch = ReduceLROnPlateau(optimizer, factor=0.1, patience=50)
-    >>> lr_sch = LearningRateScheduler(scheduler=sch, metric='MeanSquaredError', on_train=False)
+
+    The default metric is the loss. You can choose the validation or the
+    training loss or you can pass another metric by doing:
+
+    >>> lr_sch = LearningRateScheduler(
+    ...     scheduler=sch, metric='MeanSquaredError', on_train=False)
     >>> metrics = [torchmetrics.MeanSquaredError]
     >>> trainer = Trainer(callbacks=[lr_sch], metrics=metrics, **kwargs)
 
@@ -189,22 +201,33 @@ class LearningRateScheduler(Callback):
 
     def __repr__(self) -> str:
         sch = type(self.scheduler).__name__
-        return f"LearningRateScheduler(scheduler={sch}, metric={self.metric})"
+        return (
+            f"LearningRateScheduler(scheduler={sch}, metric={self.metric}), "
+            f"on_train={self.on_train}"
+        )
+
+    def on_fit_start(self, params_dict: dict) -> None:
+        accelerator = params_dict[ParamsDict.ACCELERATOR]
+        self.scheduler = accelerator.prepare(self.scheduler)
 
     def on_train_step_end(self, params_dict: dict) -> None:
         if self.metric is not None:
             key = "train" if self.on_train else "validation"
             epoch_hist = params_dict[ParamsDict.EPOCH_HISTORY]
-            metric = epoch_hist[self.metric][key][-1]
-            self.scheduler.step(metric)
+            epoch_number = params_dict[ParamsDict.EPOCH_NUMBER]
+
+            # avoid failing in first epoch when on_train=False
+            if epoch_number > 1:
+                metric = epoch_hist[self.metric][key][-1]
+                self.scheduler.step(metric)
         else:
             self.scheduler.step()
 
 
 class GPUStats(Callback):
-    """GPU stats logger. 
-    
-    The list of available queries can be found on NVIDIA smi queries. See 
+    """GPU stats logger.
+
+    The list of available queries can be found on NVIDIA smi queries. See
     `Notes` section for more information.
 
     Parameters
@@ -241,9 +264,15 @@ class GPUStats(Callback):
         self.format = format
         self.update_step = update_step
 
-        self.log_name = 'GPU Stats'
+        self.log_name = "GPU Stats"
 
-    def on_epoch_end(self, params_dict):
+    def __repr__(self) -> str:
+        return (
+            f"GPUStats(format={self.format}, queries={self.queries}, "
+            f"queries={self.queries})"
+        )
+
+    def on_epoch_end(self, params_dict: dict) -> None:
         epoch_number = params_dict[ParamsDict.EPOCH_NUMBER]
 
         if epoch_number == 1 or epoch_number % self.update_step == 0:
@@ -299,17 +328,32 @@ class RichProgressBar(Callback):
         self.prec = precision
         self.log_lr = log_lr
 
-        self.log_name = 'Rich Bar'
+        self.log_name = "Rich Bar"
+
+    def __repr__(self) -> str:
+        return (
+            f"RichProgressBar(display_step={self.display_step}, "
+            f"log_lr={self.log_lr}, precision={self.precision})"
+        )
+
+    def on_fit_start(self, params_dict: dict) -> None:
+        dev = params_dict[ParamsDict.DEVICE]
+        self.logger.info(f"Starting training process on {dev}\n")
 
     def on_train_batch_end(self, params_dict: dict) -> None:
         epoch = params_dict[ParamsDict.EPOCH_NUMBER]
+        accelerator = params_dict[ParamsDict.ACCELERATOR]
+
         if epoch % self.display_step == 0 or epoch == 1:
+            accelerator.wait_for_everyone()
             self.progress_bar.advance(self.epoch_task, 1)
 
     def on_validation_batch_end(self, params_dict: dict) -> None:
         epoch = params_dict[ParamsDict.EPOCH_NUMBER]
+        accelerator = params_dict[ParamsDict.ACCELERATOR]
+
         if epoch % self.display_step == 0 or epoch == 1:
-            # advance bar
+            accelerator.wait_for_everyone()
             self.progress_bar.advance(self.epoch_task, 1)
 
     def on_epoch_start(self, params_dict: dict) -> None:
@@ -345,9 +389,16 @@ class RichProgressBar(Callback):
         if epoch % self.display_step == 0 or epoch == 1:
             # update metrics
             text = self.render_text(params_dict[ParamsDict.EPOCH_HISTORY])
-            self.logger.info(text)
+            self.logger.info(text)  # DISC: use included Rich logger?
             self.progress_bar.stop()
 
+    def on_fit_end(self, params_dict):
+        total_time = params_dict[ParamsDict.TOTAL_TIME]
+        # final message
+        self.logger.info(
+            f"""\nEnd of training. Total time: {total_time:0.5f} seconds"""
+        )
+
     def render_text(self, update_dict):
         text_format = ""
 
@@ -358,20 +409,20 @@ class RichProgressBar(Callback):
 
                 if text_format:  # not empty
                     text_format = (
-                        f"{text_format} • {metric} > Train: "
+                        f"{text_format} • {metric} -> Train: "
                         f"{train_metric:.{self.prec}e} | "
                         f"Validation: {val_metric:.{self.prec}e}"
                     )
                 else:
                     text_format = (
-                        f"{metric} > Train: "
+                        f"{metric} -> Train: "
                         f"{train_metric:.{self.prec}e} | Validation: "
                         f"{val_metric:.{self.prec}e}"
                     )
             else:
                 if self.log_lr:
                     text_format = (
-                        f"{text_format} • LearningRate: "
+                        f"{text_format} • Learning Rate: "
                         f"{update_dict[metric][-1]}"
                     )
 
@@ -379,11 +430,11 @@ class RichProgressBar(Callback):
 
 
 class StochasticWeightAveraging(Callback):
-    """Applies a stochastic weight averaging to the training process. 
-    
-    If you were to use a learning rate scheduler in addition to stochastic 
-    averaging, you must pass both to the constructor of this class instead of 
-    creating an individual callback for the standard lr scheduler. See 
+    """Applies a stochastic weight averaging to the training process.
+
+    If you were to use a learning rate scheduler in addition to stochastic
+    averaging, you must pass both to the constructor of this class instead of
+    creating an individual callback for the standard lr scheduler. See
     `Examples` section.
 
     Parameters
@@ -416,13 +467,15 @@ class StochasticWeightAveraging(Callback):
     >>> optimizer, model, criterion = ...
     >>> swa_model = torch.optim.swa_utils.AveragedModel(model)
     >>> swa_start = 160
-    >>> scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=300)
+    >>> scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
+    ...    optimizer, T_max=300)
     >>> swa_scheduler = SWALR(optimizer, swa_lr=0.05)
-    >>> swa_callback = StochasticWeightAveraging(swa_scheduler, swa_start, scheduler=scheduler)
+    >>> swa_callback = StochasticWeightAveraging(
+    ...     swa_scheduler, swa_start, scheduler=scheduler)
     >>> trainer = Trainer(callbacks=[swa_callback], **kwargs)
     >>> history = trainer.fit(...)
 
-    Now we can the SWA model by simply calling:
+    Now we can get the SWA model by simply calling:
     >>> swa_model = swa_callback.get_swa_model()
     """
 
@@ -445,10 +498,17 @@ class StochasticWeightAveraging(Callback):
         self.__swa_model = None
 
     def __repr__(self) -> str:
-        return f"StochasticWeightAveraging(swa_scheduler={self.swa_scheduler}, start_epoch={self.start_epoch})"
+        return (
+            f"StochasticWeightAveraging(swa_scheduler={self.swa_scheduler}, "
+            f"start_epoch={self.start_epoch})"
+        )
 
     def on_fit_start(self, params_dict: dict) -> None:
         model = params_dict[ParamsDict.MODEL]
+        accelerator = params_dict[ParamsDict.ACCELERATOR]
+
+        self.scheduler = accelerator.prepare(self.scheduler)
+        self.swa_scheduler = accelerator.prepare(self.swa_scheduler)
         self.__swa_model = AveragedModel(model)
 
     def on_train_step_end(self, params_dict: dict) -> None:
@@ -489,3 +549,164 @@ class StochasticWeightAveraging(Callback):
             SWA model.
         """
         return self.__swa_model
+
+
+class L1Regularization(Callback):
+    """Applies L1 regularization over the model parameters.
+
+    L1 is usually called 'Lasso Regression' (Least Absolute Shrinkage and
+    Selection Operator). This callbacks is only applied to the train loss.
+
+    Parameters
+    ----------
+    regularization_rate : float
+        Regularization rate. Also called `lambda`.
+    biases : bool, optional, default: False
+        Whether to apply regularization over bias terms (True) or not (False).
+
+    Note
+    ----
+    The penalty term already handles the product by the lambda regularization
+    rate.
+
+    """
+
+    def __init__(self, regularization_rate: float, biases: bool = False):
+        super().__init__()
+
+        self.rate = regularization_rate
+        self.biases = biases
+
+    def on_loss_step_end(self, params_dict: dict) -> None:
+        batch_tr_loss = params_dict[ParamsDict.BATCH_TRAIN_LOSS]
+        device = params_dict[ParamsDict.DEVICE]
+        model = params_dict[ParamsDict.MODEL]
+
+        # Initialize with tensor, cannot be scalar
+        penalty_term = torch.zeros(1, 1, requires_grad=True).to(device)
+
+        for name, param in model.named_parameters():
+            if not self.biases and name.endswith("bias"):
+                continue
+
+            penalty_term = penalty_term + param.norm(p=1)
+
+        total_penalty = self.rate * penalty_term
+        loss = total_penalty + batch_tr_loss
+
+        # set loss
+        params_dict[ParamsDict.BATCH_TRAIN_LOSS] = loss
+
+
+class L2Regularization(Callback):
+    """Applies L2 regularization over the model parameters.
+
+    L2 is usually called 'Ridge Regression'. This callbacks is only applied to
+    the train loss.
+
+    Parameters
+    ----------
+    regularization_rate : float
+        Regularization rate. Also called `lambda`.
+    biases : bool, optional, default: False
+        Whether to apply regularization over bias terms (True) or not (False).
+
+    Note
+    ----
+    The penalty term already handles the product by the lambda regularization
+    rate.
+
+    """
+
+    def __init__(self, regularization_rate: float, biases: bool = False):
+        super().__init__()
+
+        self.rate = regularization_rate
+        self.biases = biases
+
+    def on_loss_step_end(self, params_dict: dict) -> None:
+        batch_tr_loss = params_dict[ParamsDict.BATCH_TRAIN_LOSS]
+        device = params_dict[ParamsDict.DEVICE]
+        model = params_dict[ParamsDict.MODEL]
+
+        # Initialize with tensor, cannot be scalar
+        penalty_term = torch.zeros(1, 1, requires_grad=True).to(device)
+
+        for name, param in model.named_parameters():
+            if not self.biases and name.endswith("bias"):
+                continue
+
+            penalty_term = penalty_term + param.norm(p=2)
+
+        total_penalty = self.rate * penalty_term
+        loss = total_penalty + batch_tr_loss
+
+        # set loss
+        params_dict[ParamsDict.BATCH_TRAIN_LOSS] = loss
+
+
+class ElasticNetRegularization(Callback):
+    r"""Linear combination of L1 and L2.
+
+    According to [1], the lasso penalty is somewhat indifferent to the choice
+    among a set of strong but correlated variables. The ridge penalty, on the
+    other hand, tends to shrink the coefficients of correlated variables toward
+    each other. Elastic net combines both using a weighting factor:
+
+    .. math::
+
+        \sum_{j=1}^{p} ( \alpha |\beta_{j}| + (1 + \alpha) \beta_{j}^{2} )
+
+    Parameters
+    ----------
+    regularization_rate : float
+        Regularization rate. Also called `lambda`.
+    alpha : float
+        Parameter to determine the mix of the penalties.
+    biases : bool, optional, default: False
+        Whether to apply regularization over bias terms (True) or not (False).
+
+    Note
+    ----
+    The penalty term already handles the product by the lambda regularization
+    rate.
+
+    References
+    ----------
+    .. [1] Trevor Hastie, Robert Tibshirani, Jerome Friedman - The Elements of
+       Statistical Learning.
+
+    """
+
+    def __init__(
+        self, regularization_rate: float, alpha: float, biases: bool = False
+    ):
+        super().__init__()
+
+        self.rate = regularization_rate
+        self.biases = biases
+        self.alpha = alpha
+
+    def on_loss_step_end(self, params_dict: dict) -> None:
+        batch_tr_loss = params_dict[ParamsDict.BATCH_TRAIN_LOSS]
+        device = params_dict[ParamsDict.DEVICE]
+        model = params_dict[ParamsDict.MODEL]
+
+        # Initialize with tensor, cannot be scalar
+        penalty_term = torch.zeros(1, 1, requires_grad=True).to(device)
+
+        for name, param in model.named_parameters():
+            if not self.biases and name.endswith("bias"):
+                continue
+
+            l1 = param.norm(p=1)
+            l2 = param.norm(p=2)
+            penalty_term = penalty_term + (
+                self.alpha * l1 + (1 - self.alpha) * l2
+            )
+
+        total_penalty = self.rate * penalty_term
+        loss = total_penalty + batch_tr_loss
+
+        # set loss
+        params_dict[ParamsDict.BATCH_TRAIN_LOSS] = loss
